from decimal import Decimal

from scipy.stats import rv_discrete

from analysis.forecasting.forecast_strategies.form_forecast_strategy import (
    FormForecastStrategy,
)
from analysis.forecasting.form_forecast import FormForecast
from analysis.horse import Horse
from analysis.race_conditions import RaceConditions


class RepeatLto(FormForecastStrategy):
    def __call__(self, horse: Horse, _race_conditions: RaceConditions) -> FormForecast:
        lto_rating = horse.performances[-1].form_rating
        probability_distribution = rv_discrete(values=([lto_rating], [1]))
        solidity = Decimal(1.0)
        return FormForecast(probability_distribution, solidity)



from decimal import Decimal

from scipy.stats import triang

from analysis.forecasting.forecast_strategies.form_forecast_strategy import (
    FormForecastStrategy,
)
from analysis.forecasting.form_forecast import FormForecast
from analysis.horse import Horse
from analysis.race_conditions import RaceConditions


class MinModeMaxTriangular(FormForecastStrategy):
    def __call__(self, horse: Horse, _race_conditions: RaceConditions) -> FormForecast:
        rated_performances = [
            p.form_rating for p in horse.performances if p.form_rating
        ]

        min_rating = min(rated_performances)
        max_rating = max(rated_performances)
        mode_rating = sum(rated_performances) / len(horse.performances)

        loc = min_rating
        scale = max_rating - min_rating
        c = (mode_rating - min_rating) / scale

        probability_distribution = triang(c, loc, scale)
        solidity = Decimal(1.0)
        return FormForecast(probability_distribution, solidity)



from typing import Callable, Self

from numpy import array, prod


class LikenessMixin:
    _likeness_functions: dict[str, Callable] = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)  # forwards all unused arguments

    def like(self, other: Self) -> float:
        likeness_base = float(bool(len(self._likeness_functions)) or self == other)

        factors = []
        for attribute, function in self._likeness_functions.items():
            try:
                a = getattr(self, attribute)
                b = getattr(other, attribute)
                factors.append(function(a, b))
            except Exception:
                raise ValueError(
                    f"Unable to calculate likeness: {attribute} comparison failed"
                )

        return prod(array([likeness_base, *factors]))


def discount(calculation: float) -> float:
    if calculation < 0:
        raise ValueError("Discount function must return a value below 1")
    return max(1 - calculation, 0)


def ignore(a: object, b: object) -> float:
    return 1


def likeness(a: LikenessMixin, b: LikenessMixin) -> float:
    return a.like(b)



from horsetalk import RaceLevel


class FlatRaceLevelParRating:
    _lookup = {
        "Group 1": 140,
        "Group 2": 130,
        "Group 3": 120,
        "Listed": 110,
        "Class 2": 100,
        "Class 3": 90,
        "Class 4": 80,
        "Class 5": 70,
        "Class 6": 60,
        "Class 7": 50,
    }

    @staticmethod
    def max(level: RaceLevel) -> int:
        if level.grade.racing_code.name == "NATIONAL_HUNT":
            raise ValueError("These race level pars are for flat racing only")
        return FlatRaceLevelParRating._lookup[str(level.grade) or str(level.class_)]

    @staticmethod
    def average(level: RaceLevel) -> int:
        return FlatRaceLevelParRating.max(level) - 35

    @staticmethod
    def min(level: RaceLevel) -> int:
        if level.grade.racing_code.name == "NATIONAL_HUNT":
            raise ValueError("These race level pars are for flat racing only")
        return FlatRaceLevelParRating.max(level) - 70



from decimal import Decimal

from .form_prediction_factor import FormPredictionFactor


class ThreeYearOldVsOlderBonus(FormPredictionFactor):
    def _calc(self) -> Decimal:
        return Decimal(10 if self.horse.age == 3 else 0)



from decimal import Decimal
from typing import Self

from horsetalk import HorseAge, Horselength
from measurement.measures import Weight  # type: ignore
from pendulum.datetime import DateTime


class RateableRun:
    def __init__(
        self,
        beaten_distance: Horselength,
        weight_carried: Weight,
        weight_allowance: Weight,
        age: HorseAge,
    ):
        self.beaten_distance = beaten_distance
        self.weight_carried = weight_carried
        self.weight_allowance = weight_allowance
        self.age = age

    @classmethod
    def parse(
        cls,
        beaten_distance: str,
        weight_carried: str,
        weight_allowance: str,
        age: str,
        date: DateTime | None = None,
    ) -> Self:
        stones, lbs = weight_carried.split("-")
        horse_age = (
            HorseAge(int(age), context_date=date) if date else HorseAge(int(age))
        )
        return cls(
            beaten_distance=Horselength(beaten_distance),
            weight_carried=Weight(stone=Decimal(stones)) + Weight(lb=Decimal(lbs)),
            weight_allowance=Weight(lb=Decimal(weight_allowance)),
            age=horse_age,
        )



from .form_rating import FormRating
from .jockey_rating import JockeyRating
from .rating import Rating
from .time_rating import TimeRating

__all__ = [
    "FormRating",
    "JockeyRating",
    "Rating",
    "TimeRating",
]



from decimal import Decimal
from typing import Self, Tuple, Type

import numpy as np
from horsetalk import Horselength


class BeatenDistances(tuple):
    """
    This class represents a tuple of horselength values used for calculating the distances between horses in a race.

    Args:
        value (tuple[int | float | Decimal | str]): A tuple containing the horselength values.

    Raises:
        TypeError: If an attempt is made to directly instantiate the BeatenDistances class.
        ValueError: If the first element of value is not None or 0 or if an invalid horselength value is encountered in value.

    Returns:
        An instance of the BeatenDistances class.
    """

    def __new__(cls: Type[Self], values: Tuple[int | float | Decimal | str]) -> Self:
        if cls == BeatenDistances:
            raise TypeError("BeatenDistances cannot be instantiated directly")

        if values[0] and int(values[0]) != 0:
            raise ValueError("First element of BeatenDistances must be None or 0")

        horselengths = []
        for v in values:
            hl = Horselength(v)
            if hl < 0:
                raise ValueError(f"Negative horselength value for BeatenDistances: {v}")
            horselengths.append(hl)

        return super().__new__(cls, horselengths)  # type: ignore


class CumulativeBeatenDistances(BeatenDistances):
    """
    This class represents a tuple of horselength values used for calculating the distances between horses in a race,
    where the values represent the cumulative beaten distances.

    Args:
        value (tuple[int | float | Decimal | str]): A tuple containing the horselength values.

    Raises:
        ValueError: If any item in value is less than the previous one or if the first element of value is not None or 0.

    Returns:
        An instance of the CumulativeBeatenDistances class.
    """

    def __new__(cls: Type[Self], values: Tuple[int | float | Decimal | str]) -> Self:
        horselengths = super().__new__(cls, values)  # type: ignore

        for i, v in enumerate(horselengths[1:]):
            if v < horselengths[i]:
                raise ValueError(
                    f"Invalid value at index {i + 1}: CumulativeBeatenDistances must increase monotonically"
                )

        return horselengths  # type: ignore

    def to_marginal(self) -> "MarginalBeatenDistances":
        """
        Convert the instance to a MarginalBeatenDistances instance.

        Returns:
            MarginalBeatenDistances: The instance converted to a MarginalBeatenDistances instance.
        """
        return MarginalBeatenDistances(np.diff(self, prepend=0))  # type: ignore


class MarginalBeatenDistances(BeatenDistances):
    """
    This class represents a tuple of horselength values used for calculating the distances between horses in a race,
    where the values represent the marginal beaten distances.

    Returns:
        An instance of the MarginalBeatenDistances class.
    """

    def to_cumulative(self) -> CumulativeBeatenDistances:
        """
        Convert the instance to a CumulativeBeatenDistances instance.

        Returns:
            CumulativeBeatenDistances: The instance converted to a CumulativeBeatenDistances instance.
        """
        return CumulativeBeatenDistances(np.cumsum(self))  # type: ignore



import re
from decimal import Decimal
from typing import Self

from .rating import Rating


class JockeyRating(Rating):
    @classmethod
    def parse(cls, string: str) -> Self:
        """
        Parses a string into a JockeyRating instance.

        Args:
            string: The string to parse.

        Returns:
            A new JockeyRating instance.
        """
        match = re.match(r"(\d+)(\**)", string)

        if not match:
            raise ValueError(f"Could not parse JockeyRating from '{string}'")

        number, stars = match.groups()
        return cls(number, solidity=Decimal(1 - len(stars) * 0.25))



from decimal import Decimal
from typing import Self

from horsetalk import Horselength
from measurement.measures import Weight  # type: ignore

from .rating import Rating


class FormRating(Rating):
    def adjust(self, delta: float | Decimal | str) -> Self:
        """
        Returns a new FormRating instance with the given delta added to the value.

        Args:
            delta: The delta to add to the value.

        Returns:
            A new FormRating instance.
        """
        return self.__class__(self + Decimal(delta), solidity=self.solidity)

    @classmethod
    def calculate(
        cls,
        *,
        beaten_distance: Horselength,
        weight_per_length: Weight,
        weight_differential: Weight = Weight(lb=0),
        weight_for_age: Weight = Weight(lb=0),
        baseline: Decimal = Decimal(0),
    ) -> Self:
        """
        Calculates the FormRating for a horse from its beaten distance and a given weight per length

        Args:
            beaten_distance: The beaten distance to use in the calculation.
            weight_per_length: The weight per length to use in the calculation.
            weight_differential: The weight adjustment, e.g due to weight carried/jockey allowance, to use in the calculation.
            weight_for_age: The weight for age allowance to use in the calculation.
            baseline: The baseline rating to use in the calculation.

        Returns:
            The FormRating for the horse.
        """
        raw_delta = beaten_distance * round(Decimal(weight_per_length.lb), 2)
        weight_adjustment = round(Decimal((weight_differential + weight_for_age).lb), 2)
        value = baseline - raw_delta + weight_adjustment
        solidity = 1 - (raw_delta.sqrt()) / 10
        return cls(value, solidity=solidity)



from analysis.forecasting.form_forecast import FormForecast
from analysis.horse import Horse
from analysis.race_conditions import RaceConditions


class FormForecastStrategy:
    def __call__(self, horse: Horse, race_conditions: RaceConditions) -> FormForecast:
        raise NotImplementedError



from decimal import Decimal
from typing import List

from analysis.rating import FormRating

from .form_rating_level_setting_strategy import FormRatingLevelSettingStrategy


class RateThroughRunnerStrategy(FormRatingLevelSettingStrategy):
    """
    A strategy for setting the level of a form rating.
    """

    def __init__(self, index: int = 0):
        """
        Initialize a RateThroughRunnerStrategy instance.

        Args:
            index: The index of the runner through which to rate the race.
        """
        self.index = index

    def _adjustment(self, priors: List[FormRating], raw: List[FormRating]) -> Decimal:
        """
        Calculate the adjustment to apply to the raw ratings.

        Args:
            priors: The prior assumptions about what the ratings should be.
            raw: The raw ratings.

        Returns:
            The adjustment to apply to the raw ratings.
        """
        return Decimal(priors[self.index] - raw[self.index])



from decimal import Decimal
from typing import List

from analysis.rating import FormRating

from .form_rating_level_setting_strategy import FormRatingLevelSettingStrategy


class LeastDifferenceStrategy(FormRatingLevelSettingStrategy):
    """
    A strategy for setting the level of a form rating.
    """

    def __init__(self, num_runners_to_use: int | None = None):
        """
        Initialize a LeastDifferenceStrategy instance.

        Args:
            num_runners_to_use: The number of runners to use in the calculation. If None, use all
                runners.
        """
        self.num = num_runners_to_use

    def _adjustment(self, priors: List[FormRating], raw: List[FormRating]) -> Decimal:
        """
        Calculate the adjustment to apply to the raw ratings.

        Args:
            priors: The prior assumptions about what the ratings should be.
            raw: The raw ratings.

        Returns:
            The adjustment to apply to the raw ratings.
        """
        useable_priors = priors[: self.num] if self.num else priors
        useable_raw = raw[: self.num] if self.num else raw
        ave_prior = sum(useable_priors) / len(useable_priors)
        ave_raw = sum(useable_raw) / len(useable_raw)
        return Decimal(ave_prior) - Decimal(ave_raw)



from decimal import Decimal

from analysis.weight_for_age_converter import WeightForAgeConverter


class WeightForAgeScale(WeightForAgeConverter):
    """
    A class that represents a weight for age scale, which is a dictionary of dictionaries of Weight objects.
    The outer dictionary is keyed by distance, and the inner dictionary is keyed by age range.
    Values within an age range will be interpolated linearly based on the age of the horse.

    Inherits from WeightForAgeConverter and adds the ability to lookup weights from a table based on age, distance, and date.
    """

    def _process(self, key: range, val: int, age_in_days: int) -> Decimal:
        proportion = (key[-1] - age_in_days) / len(key)
        allowance = proportion * val
        return Decimal(str(allowance))



from decimal import Decimal

from .form_prediction_factor import FormPredictionFactor


class ExperienceImprovementBonus(FormPredictionFactor):
    def _calc(self) -> Decimal:
        perfs = self.horse.performances

        if not perfs:
            return Decimal(0)

        return Decimal(max(6 - len(perfs), 0) ** 3 / 10)



import random
from decimal import Decimal
from itertools import starmap
from typing import Callable

from pybet import Odds  # type: ignore


class MonteCarloSimulator:
    def __init__(self, selection_function: Callable, runner_data: list[list]):
        self.selection_function = selection_function
        self.runner_data = runner_data
        n = len(runner_data)
        self.result_grid = [{(k + 1): 0 for k in range(n)} for _ in range(n)]
        self._n = n
        self._validate_runner_data()

    def run(self, times: int = 1000):
        for _ in range(times):
            result = list(starmap(self.selection_function, self.runner_data))
            randomiser = [random.random() for _ in self.runner_data]
            # add a small random factor to separate equal values
            result = [r + randomiser[i] for i, r in enumerate(result)]
            ordering = [sorted(result, reverse=True).index(r) for r in result]
            for i, o in enumerate(ordering):
                self.result_grid[i][o + 1] += 1
        return [
            {k: Odds.probability(Decimal(str(v / times))) for k, v in row.items()}
            for row in self.result_grid
        ]

    def _validate_runner_data(self):
        for datum in self.runner_data:
            self.selection_function(*datum)



from decimal import Decimal
from typing import Any

import numpy as np
from horsetalk import Horselength

from analysis.flat_race_level_par_rating import FlatRaceLevelParRating
from analysis.horse import Horse
from analysis.race_conditions import RaceConditions
from analysis.rating import FormRating

from .forecast_strategies import (
    ExperienceImprovementBonus,
    HandicapDebutBonus,
    ThreeYearOldVsOlderBonus,
)


def adjusted_win_rating(
    win_rating: FormRating,
    lengths_ahead: Horselength,
    lbs_per_length: float = 4,
    comment: str = "",
) -> FormRating:
    bonus_words = ["readily", "easily", "comfortably", "impressive"]
    adjustment = np.prod(
        np.array(
            [
                Decimal(x)
                for x in [
                    lengths_ahead,
                    (lbs_per_length**0.5),
                    (
                        1.5
                        if any(word in bonus_words for word in comment.split())
                        else 1
                    ),
                ]
            ]
        )
    )
    return win_rating.adjust(adjustment)


class FormPredictor:
    """
    A class for predicting form ratings for a race using current race conditions and past performance.
    """

    def __init__(
        self,
        horse: Horse,
        race_conditions: RaceConditions,
    ):
        self.horse = horse
        self.race_conditions = race_conditions

    @property
    def _ceiling(self) -> FormRating | None:
        """
        The ceiling of the past performances.

        Returns:
            The ceiling of the past performances.

        """
        if not len(self._rated_performances):
            return None

        top_performance = max(
            list(self._rated_performances), key=lambda x: x.form_rating
        )
        return (
            top_performance.form_rating
            if not (top_performance.is_win and len(self._rated_performances) <= 4)
            else None
        )

    @property
    def _additions(self) -> float:
        """
        Additions to make to form rating based on various potential improvements.

        Returns:
            Additions to make to form rating based on various potential improvements.

        """
        bonuses = [
            ExperienceImprovementBonus,
            HandicapDebutBonus,
            ThreeYearOldVsOlderBonus,
        ]
        return float(sum(b(self.horse, self.race_conditions) for b in bonuses))

    @property
    def _floor(self) -> FormRating | None:
        """
        The floor of the past performances.

        Returns:
            The floor of the past performances.

        """
        if not len(self._rated_performances):
            return None

        worst_performance = min(
            list(self._rated_performances),
            key=lambda x: x.form_rating,
        )
        return r if (r := worst_performance.form_rating).solidity > 0 else None

    @property
    def _rated_performances(self) -> np.ndarray[float, np.dtype[Any]]:
        """
        A numpy array containing past performances that have been rated.

        Returns:
            A numpy array containing past performances that have been rated.
        """
        return np.array([p for p in self.horse.performances if p and p.form_rating])

    @property
    def _ratings(self) -> np.ndarray[float, np.dtype[Any]]:
        """
        A numpy array containing the ratings of the past performances.

        Returns:
            A numpy array containing the ratings of the past performances.

        """
        return np.array([float(p.form_rating) for p in self._rated_performances])

    @property
    def _solidity(self) -> Decimal:
        """
        The solidity of the past performances.

        Returns:
            The solidity of the past performances.

        """
        return Decimal(str(min(sum(self._weightings), 1)))

    @property
    def _weightings(self) -> np.ndarray[float, np.dtype[Any]]:
        """
        A numpy array containing the weightings of the past performances.

        Returns:
            A numpy array containing the weightings of the past performances.

        """
        return np.array(
            [
                float(self.race_conditions.like(key))
                * float(value.form_rating.solidity)
                for key, value in self.horse.race_career.items()
                if value and value.form_rating
            ]
        )

    def _weighted_percentile(
        self, percentile=5, zero_val=0, *, inverted=False
    ) -> np.ndarray[Any, Any]:
        """
        Calculate the weighted percentile of the past performances.

        Args:
            percentile (int, optional): The percentile to calculate. Defaults to 5.
            zero_val (int, optional): The value to use for the zeroth percentile. Defaults to 0.
            inverted (bool, optional): Whether to invert the ordering of the past performances. Defaults to False.

        Returns:
            np.ndarray[Any, Any]: The weighted percentile of the past performances.

        """
        indices = np.argsort(self._ratings)[:: -1 if inverted else 1]
        ordered_ratings = self._ratings[indices]
        ordered_weightings = self._weightings[indices]
        cum_percents = ordered_weightings.cumsum() / ordered_weightings.sum() * 100
        return np.interp(percentile, [0, *cum_percents], [zero_val, *ordered_ratings])

    def predict(self) -> FormRating | None:
        """
        Predict the expected FormRating for a horse, given past performances and current conditions

        Returns:
            FormRating: The predicted FormRating for the horse.
        """

        if not len(self._ratings):
            return FormRating(
                FlatRaceLevelParRating.average(self.race_conditions.race_level)
            )

        average = np.average(
            [
                float(
                    adjusted_win_rating(
                        p.form_rating, abs(min(p.beaten_distance or 0, 0))
                    )
                )
                for p in self._rated_performances
            ],
            weights=self._weightings,
        )  # type: ignore

        return FormRating(average + self._additions, solidity=self._solidity)

    def lower_bound(self):
        """
        Predict the lowest possible FormRating for a horse, given past performances and current conditions

        Args:
            min_val (int, optional): The minimum value to use for the zeroth percentile. Defaults to 0.

        Returns:
            FormRating: The lowest possible FormRating for the horse.
        """
        race_min_val = FlatRaceLevelParRating.min(self.race_conditions.race_level)

        if not len(self._ratings):
            return race_min_val * 1.25

        horse_min_val = (
            float(self._floor) + self._additions if self._floor else race_min_val
        )

        zero_value = float(race_min_val * (1 - self._solidity)) + (
            horse_min_val * float(self._solidity)
        )

        return (
            self._weighted_percentile(zero_val=zero_value)
            if self._floor
            else (0.05 * (min(self._ratings) - race_min_val) + race_min_val)
        )

    def upper_bound(self):
        """
        Predict the highest possible FormRating for a horse, given past performances and current conditions

        Args:
            max_val (int, optional): The maximum value to use for the zeroth percentile. Defaults to 140.

        Returns:
            FormRating: The highest possible FormRating for the horse.

        """
        race_max_val = FlatRaceLevelParRating.max(self.race_conditions.race_level)

        if not len(self._ratings):
            return race_max_val * 0.8

        horse_max_val = (
            float(self._ceiling) + self._additions if self._ceiling else race_max_val
        )

        zero_value = float(race_max_val * (1 - self._solidity)) + (
            horse_max_val * float(self._solidity)
        )
        return self._weighted_percentile(zero_val=zero_value, inverted=True)

    def to_debug(self):
        """
        Return a dictionary of the attributes of the FormPredictor for debugging purposes.

        Returns:
            dict: A dictionary of the attributes of the FormPredictor for debugging purposes.
        """
        return {
            "ceiling": self._ceiling,
            "additions": self._additions,
            "floor": self._floor,
            "ratings": self._ratings,
            "solidity": self._solidity,
            "performance_summary": [
                {
                    "course": k.racecourse.name,
                    "date": k.datetime.datetime.format("YYMMDD"),
                    "rating": v.form_rating,
                    "likeness": float(self.race_conditions.like(k)),
                    "solidity": float(v.form_rating.solidity),
                }
                for k, v in self.horse.race_career.items()
                if v and v.form_rating
            ],
        }



from pendulum import DateTime

from .likeness import LikenessMixin, discount


class RaceDateTime(LikenessMixin):
    _likeness_functions = {
        # date likeness reaches 0 after 1000 days
        "datetime": lambda a, b: discount((abs(a - b).days / 1000) ** 0.5)
    }

    def __init__(self, datetime: DateTime):
        self.datetime = datetime

    def __eq__(self, other: object):
        if not isinstance(other, RaceDateTime):
            return NotImplemented

        return self.datetime == other.datetime

    def __lt__(self, other: "RaceDateTime"):
        return self.datetime < other.datetime

    def __gt__(self, other: "RaceDateTime"):
        return self.datetime > other.datetime

    def format(self, fmt: str) -> str:
        return self.datetime.format(fmt)



from decimal import Decimal
from typing import List

from pybet import Odds  # type: ignore

from analysis.rating import FormRating


class RatingsToOddsConverter:
    @staticmethod
    def convert(ratings: List[FormRating], alpha: Decimal = Decimal(1)) -> List[Odds]:
        weightings = [alpha**rating for rating in ratings]
        total = sum(weightings)
        return [Odds.probability(weighting / total) for weighting in weightings]



from dataclasses import dataclass
from decimal import Decimal

from scipy.stats import rv_continuous, rv_discrete


@dataclass(frozen=True)
class FormForecast:
    """
    A forecast of a horse's form rating for a given set of conditions
    """

    probability_distribution: rv_discrete | rv_continuous
    solidity: Decimal



from horsetalk import RacePerformance as HorsetalkRacePerformance

from analysis.rating import FormRating, TimeRating


class RacePerformance(HorsetalkRacePerformance):
    def __init__(
        self,
        *args,
        form_rating: FormRating | None = None,
        time_rating: TimeRating | None = None,
        **kwargs,
    ):
        self.form_rating = form_rating
        self.time_rating = time_rating
        super().__init__(*args, **kwargs)



from itertools import starmap
from typing import List, Self

from horsetalk import RaceDistance
from pendulum import duration
from pendulum.datetime import DateTime
from pendulum.duration import Duration

from analysis.rateable_run import RateableRun


class RateableResult:
    def __init__(
        self,
        datetime: DateTime,
        distance: RaceDistance,
        win_time: Duration,
        runs: List[RateableRun],
    ):
        self.datetime = datetime
        self.distance = distance
        self.win_time = win_time
        self.runs = runs

    @classmethod
    def parse(
        cls,
        datetime: DateTime,
        distance: str,
        win_time: str,
        runs: List[List[str | DateTime | None]],
    ) -> Self:
        mins, secs = win_time.split(":")
        return cls(
            datetime=datetime,
            distance=RaceDistance(distance),
            win_time=duration(minutes=int(mins), seconds=float(secs)),
            runs=list(starmap(RateableRun.parse, runs)),  # type: ignore
        )



from horsetalk import Racecourse as HorsetalkRacecourse

from .likeness import LikenessMixin


def enum_likeness(step: float):
    return lambda a, b: max(
        a == b, 1 - step if any(x.value == 0 for x in [a, b]) else 1 - step * 2
    )


class Racecourse(LikenessMixin, HorsetalkRacecourse):
    """
    A class to represent a racecourse, wrapping the horsetalk class of the same name.

    """

    _likeness_functions = {
        "handedness": enum_likeness(0.1667),
        "contour": enum_likeness(0.1),
        "shape": enum_likeness(0.1),
        "style": enum_likeness(0.1),
        "surface": lambda a, b: max(
            a == b,
            0.83
            if all(x.name in ["FIBRESAND", "POLYTRACK", "TAPETA"] for x in [a, b])
            else 0.67,
        ),
    }



from analysis import Horse, RaceConditions
from analysis.forecasting.forecast_strategies import FormForecastStrategy
from analysis.forecasting.form_forecast import FormForecast


class FormForecaster:
    def __init__(self, strategy: FormForecastStrategy):
        self.strategy = strategy

    def forecast(
        self,
        horse: Horse,
        race_conditions: RaceConditions,
    ) -> FormForecast:
        return self.strategy(horse, race_conditions)



from decimal import Decimal
from typing import Self


class Rating(Decimal):
    """
    A Decimal subclass that represents ratings. To be used as parent class.
    """

    solidity: Decimal

    def __new__(
        cls,
        value: float | Decimal | str,
        *,
        solidity: float | Decimal | str = 1,
    ) -> Self:
        """
        Creates a new Rating instance.

        Args:
            value: The value to create the Rating instance from.
            solidity: The solidity (i.e. reliability) of the rating on a scale from 0 to 1.

        Returns:
            A new Rating instance.
        """
        instance = super().__new__(cls, value)
        instance.solidity = Decimal(solidity)
        return instance



from horsetalk import RaceDistance as HorsetalkRaceDistance

from .likeness import LikenessMixin, discount


class RaceDistance(LikenessMixin, HorsetalkRaceDistance):
    _likeness_functions = {
        "furlong": lambda a, b: discount((float(abs(a - b) / min(a, b)) / 1.5) ** 0.5)
    }

    def __getattr__(self, name):
        if name == "like":
            return self.like

        return super().__getattr__(name)



from decimal import Decimal
from typing import Self

from analysis.horse import Horse
from analysis.race_conditions import RaceConditions


class FormPredictionFactor(Decimal):
    def __init__(self, horse: Horse, race_conditions: RaceConditions):
        self.horse = horse
        self.race_conditions = race_conditions
        self._value = self._calc()

    def __new__(cls, *args, **kwargs) -> Self:
        instance = super().__new__(cls)
        instance.__init__(*args, **kwargs) # type: ignore
        return super().__new__(cls, instance._value)

    def _calc(self) -> Decimal:
        raise NotImplementedError



from decimal import Decimal

from horsetalk import RaceDesignation

from .form_prediction_factor import FormPredictionFactor


class HandicapDebutBonus(FormPredictionFactor):
    def _calc(self) -> Decimal:
        return Decimal(
            5
            if self.horse.is_handicap_debutant
            and self.race_conditions.race_designation == RaceDesignation.HANDICAP
            else 0
        )



from horsetalk import RaceConditions as HorsetalkRaceConditions

from .likeness import LikenessMixin, ignore, likeness
from .race_datetime import RaceDateTime


class RaceConditions(LikenessMixin, HorsetalkRaceConditions):
    _likeness_functions = {
        "datetime": likeness,
        "racecourse": likeness,
        "distance": likeness,
        "going": likeness,
        "race_designation": ignore,
        # lambda a, b: 1 if a == b else 0.83,
        "race_level": ignore,
        # lambda a, b: int(a == b)
        # or discount(abs(a.class_.value - b.class_.value) * 0.05),
    }

    def __init__(self, *args, **kwargs):
        kwargs["datetime"] = RaceDateTime(kwargs["datetime"])
        super().__init__(*args, **kwargs)



from decimal import Decimal
from typing import List

from analysis.rating import FormRating


class FormRatingLevelSettingStrategy:
    """
    A strategy for setting the level of a form rating.
    """

    def __call__(
        self, priors: List[FormRating], raw: List[FormRating]
    ) -> List[FormRating]:
        """
        Set the level of each form rating in the raw ratings based on the priors.

        Args:
            priors: The prior assumptions about what the ratings should be.
            raw: The raw ratings.

        Returns:
            The adjusted ratings.
        """
        if len(priors) != len(raw):
            raise ValueError("Priors and raw ratings must be the same length.")
        adjustment = self._adjustment(priors, raw)
        return [rating.adjust(adjustment) for rating in raw]

    def _adjustment(self, priors: List[FormRating], raw: List[FormRating]) -> Decimal:
        raise NotImplementedError



from typing import Any, Callable, Type

from horsetalk import RaceDistance
from measurement.measures import Weight  # type: ignore
from pendulum.duration import Duration


class WeightPerLengthFormula:
    def __init__(
        self,
        input_type: Type[RaceDistance | Duration],
        formula: Callable[[Any], Weight],
        weight_unit: str = "lb",
    ):
        self.input_type = input_type
        self.unit = weight_unit
        self._formula = formula

    def __call__(self, value: RaceDistance | Duration) -> Weight:
        if not isinstance(value, self.input_type):
            raise TypeError(f"Expected {self.input_type}, got {type(value)}")
        return Weight(**{self.unit: self._formula(value)})


BHA_GRAPH_LBS_PER_LENGTH = WeightPerLengthFormula(
    RaceDistance,
    lambda x: {
        5: 3.4,
        6: 2.8,
        7: 2.4,
        8: 2.1,
        9: 1.9,
        10: 1.75,
        11: 1.6,
        12: 1.5,
        13: 1.4,
        14: 1.3,
        15: 1.2,
        16: 1.1,
        17: 1.075,
        18: 1.05,
        19: 1.025,
        20: 1,
    }[int(x.furlong)],
    weight_unit="lb",
)



from horsetalk import RaceDistance
from measurement.measures import Weight  # type: ignore
from pendulum.duration import Duration

from analysis.rateable_result import RateableResult
from analysis.rateable_run import RateableRun
from analysis.rating import FormRating
from analysis.weight_for_age_converter import WeightForAgeConverter
from analysis.weight_for_age_scale import WeightForAgeScale
from analysis.weight_per_length_formula import WeightPerLengthFormula


class FormRater:
    """
    A class for calculating form ratings for a race using cumulative beaten distances and selected other variables.

    Attributes:
        _weight_per_length_formula: The formula to calculate the weight equivalent of being beaten by a length.
        _weight_for_age_converter: The converter to adjust weights for horse age.

    Methods:
        rate(): Calculate the form ratings for each horse in a race.
    """

    def __init__(
        self,
        weight_per_length_formula: WeightPerLengthFormula = WeightPerLengthFormula(
            RaceDistance, lambda x: 3
        ),
        weight_for_age_converter: WeightForAgeConverter = WeightForAgeScale(),
    ):
        """
        Initialize a new FormRater instance.

        Args:
            **kwargs: Optional keyword arguments:
                weight_per_length_formula: The formula for converting distance to weight per length. Default returns Weight(lb=3)
                weight_for_age_converter: The converter for calculating weight for age. Default returns Weight(lb=0)
        """
        self._weight_per_length_formula = weight_per_length_formula
        self._weight_for_age_converter = weight_for_age_converter

    def rate(self, result: RateableResult) -> tuple[FormRating, ...]:
        """
        Calculate the form ratings for each horse in the race.

        Args:
            result: The result to be rated.

        Returns:
            Tuple[FormRating, ...]: A tuple of FormRating instances representing the ratings for each horse in the race.
        """
        base_weight = self._weight_differential(result.runs[0])
        return tuple(
            FormRating.calculate(
                beaten_distance=run.beaten_distance,
                weight_per_length=self.weight_per_length(result),
                weight_differential=self._weight_differential(run, base_weight),
                weight_for_age=self._weight_for_age_converter.lookup(
                    result.distance, run.age
                ),
            )
            for run in result.runs
        )

    def weight_per_length(self, result: RateableResult) -> Weight:
        """
        Calculate the weight equivalent of being beaten by a length.

        Args:
            result: The result to be rated.

        Returns:
            Weight: The weight equivalent of being beaten by a length.
        """
        input_type = self._weight_per_length_formula.input_type
        wpl_input: RaceDistance | Duration = (
            result.distance if input_type == RaceDistance else result.win_time
        )
        return self._weight_per_length_formula(wpl_input)

    def _weight_differential(
        self, run: RateableRun, base_weight: Weight = Weight(lb=0)
    ) -> Weight:
        return run.weight_carried + run.weight_allowance - base_weight



from decimal import Decimal

from .race_distance import RaceDistance
from .weight_for_age_converter import WeightForAgeConverter


class WeightForAgeTable(WeightForAgeConverter):
    """
    A class that represents a weight for age table, such as may be found issued by a horseracing authority or form analyst.

    Inherits from WeightForAgeConverter and adds the ability to lookup weights from a table based on age, distance, and date.
    """

    def _process(self, key: range, val: int, age_in_days: int) -> Decimal:
        return Decimal(str(val))

    @classmethod
    def create(
        cls,
        row_headers: list[RaceDistance],
        column_headers: list[range],
        data: list[list[int]],
    ):
        """
        Creates a new WeightForAgeTable from a list of row headers, column headers, and data.

        Args:
            row_headers: A list of race distances.
            column_headers: A list of age-in-days ranges.
            data: A list of lists of weight allowances.

        Returns:
            A new WeightForAgeTable object.
        """
        table = WeightForAgeTable()
        if len(row_headers) != len(data):
            raise ValueError(
                "The number of row headers must equal the number of data rows."
            )

        for i, row in enumerate(data):
            if len(row) != len(column_headers):
                raise ValueError(
                    f"The number of data columns for row {i} must equal the number of row headers."
                )
            table.add(
                row_headers[i],
                {column_headers[j]: col for j, col in enumerate(row)},
            )
        return table



from .least_difference_strategy import LeastDifferenceStrategy
from .rate_through_runner_strategy import RateThroughRunnerStrategy

__all__ = [
    "LeastDifferenceStrategy",
    "RateThroughRunnerStrategy",
]



from collections import defaultdict
from decimal import Decimal
from typing import Dict, Self, Type

from horsetalk import HorseAge, RaceDistance
from measurement.measures import Weight  # type: ignore
from pendulum.datetime import DateTime


class WeightForAgeConverter(defaultdict):
    """
    The base class for all WeightForAgeConverter types, such as WeightForAgeScale and WeightForAgeTable.
    """

    def __init__(self, *, use_actual_age: bool = False) -> None:
        """
        Initialize a new instance of the WeightForAgeConverter class.
        """
        self.use_actual_age = use_actual_age
        super().__init__(lambda: defaultdict(int))

    def __new__(cls: Type[Self], *, use_actual_age: bool = False) -> Self:
        if cls == WeightForAgeConverter:
            raise TypeError("WeightForAgeConverter cannot be instantiated directly")

        return super().__new__(cls, use_actual_age)

    def add(self, distance: RaceDistance, values: Dict[range, int]) -> None:
        """
        Adds a new set of weight allowances for a given distance.

        Args:
            distance: A RaceDistance object representing a race distance in furlongs.
            values: A dictionary of weight allowances for a given age range.
        """
        self[distance.furlong] = values

    def lookup(
        self, distance: RaceDistance, age: HorseAge | int, date: DateTime | None = None
    ) -> Weight:
        """
        Looks up a weight for the given age, distance, and date.

        Args:
            distance: A RaceDistance object representing the race distance in furlongs.
            horse_age: A HorseAge object representing the age of the horse being assessed.

        Returns:
            A Weight object representing the weight for age for the given horse's age at the given distance.
            If no weight is found, a default value of zero is given.
        """
        age = HorseAge(age, context_date=date) if isinstance(age, int) else age
        days = age.actual.days if self.use_actual_age else age.official.days

        if self[distance.furlong]:
            for key, val in self[distance.furlong].items():
                if days in key:
                    return Weight(lb=self._process(key, val, days))
        return Weight(lb=0)

    def _process(self, key: range, val: int, age_in_days: int) -> Decimal:
        """
        A lambda function that is used to calculate the weight allowance for a horse's age.

        Args:
            key: A range object representing the range of days for which the weight allowance is valid.
            val: An integer representing the weight allowance for the given range.
            age_in_days: An integer representing the age of the horse in days.

        Returns:
            An integer representing the weight allowance for the given horse's age.
        """
        raise NotImplementedError("Method must be implemented by subclass")



from .experience_improvement_bonus import ExperienceImprovementBonus
from .form_forecast_strategy import FormForecastStrategy
from .handicap_debut_bonus import HandicapDebutBonus
from .three_year_old_vs_older_bonus import ThreeYearOldVsOlderBonus

__all__ = [
    "ExperienceImprovementBonus",
    "FormForecastStrategy",
    "HandicapDebutBonus",
    "ThreeYearOldVsOlderBonus",
]



from .beaten_distances import (
    BeatenDistances,
    CumulativeBeatenDistances,
    MarginalBeatenDistances,
)
from .flat_race_level_par_rating import FlatRaceLevelParRating
from .form_rater import FormRater
from .going import Going
from .horse import Horse
from .monte_carlo_simulator import MonteCarloSimulator
from .race_conditions import RaceConditions
from .race_distance import RaceDistance
from .race_performance import RacePerformance
from .racecourse import Racecourse
from .rateable_result import RateableResult
from .rateable_run import RateableRun
from .ratings_to_odds_converter import RatingsToOddsConverter
from .weight_for_age_converter import WeightForAgeConverter
from .weight_for_age_scale import WeightForAgeScale
from .weight_for_age_table import WeightForAgeTable
from .weight_per_length_formula import WeightPerLengthFormula

__all__ = [
    "BeatenDistances",
    "CumulativeBeatenDistances",
    "MarginalBeatenDistances",
    "FlatRaceLevelParRating",
    "FormRater",
    "Going",
    "Horse",
    "MonteCarloSimulator",
    "Racecourse",
    "RaceConditions",
    "RaceDistance",
    "RacePerformance",
    "RateableResult",
    "RateableRun",
    "RatingsToOddsConverter",
    "WeightForAgeConverter",
    "WeightForAgeScale",
    "WeightForAgeTable",
    "WeightPerLengthFormula",
]



from horsetalk import Going as HorsetalkGoing

from .likeness import LikenessMixin, discount

LIKENESS_DIFFERENCE_PER_POINT = 0.125


class Going(LikenessMixin, HorsetalkGoing):
    """
    A class to represent a going, wrapping the horsetalk class of the same name.
    """

    _likeness_functions = {
        "value": lambda a, b: discount(abs(a - b) * LIKENESS_DIFFERENCE_PER_POINT)
    }



from horsetalk import HorseAge, RaceDesignation

from .race_conditions import RaceConditions
from .race_performance import RacePerformance


class Horse:
    def __init__(
        self,
        name: str,
        country: str | None = None,
        age: HorseAge | None = None,
        race_career: dict[RaceConditions, RacePerformance] = {},
    ):
        self.name = name
        self.country = country
        self.age = age
        self.race_career = race_career

    @property
    def race_career(self):
        return self._race_career

    @race_career.setter
    def race_career(self, value: dict[RaceConditions, RacePerformance]):
        self._race_career = dict(sorted(value.items(), key=lambda x: x[0].datetime))

    @property
    def is_handicap_debutant(self):
        return not any(
            race.race_designation == RaceDesignation.HANDICAP
            for race in self.race_career
        )

    @property
    def is_lto_winner(self):
        return self.performances[-1].is_win

    @property
    def is_maiden(self):
        return not any(run.is_win for run in self.performances)

    @property
    def is_unbeaten(self):
        return all(run.is_win for run in self.performances)

    @property
    def performances(self):
        return list(self.race_career.values())
